/*
** Module   : SchemaWorker.java
** Abstract : Exposes schema analysis capability to pattern engine
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20050321   @20999 Created initial version. Detects and records
**                           natural table relation data among Progress
**                           schema tables.
** 002 ECF 20050526   @21290 Changes to support new expression engine
**                           implementation. Changed library registration
**                           mechanism based on superclass' modifications
**                           to support single library per pattern worker
**                           limit.
** 003 ECF 20050613   @21477 Reworked to handle adding inverse end of each
**                           detected join. Also added method to enable
**                           dumping all detected joins at the end of a
**                           source scan.
** 004 GES 20050706   @21669 Minor signature change for a utility method.
** 005 ECF 20051005   @22983 Update to store data to .schema file instead
**                           of .dict file. Necessitated by change to
**                           schema fixups ruleset, which preserves
**                           original AST as .dict and saves fixed up file
**                           with a .schema extension.
** 006 ECF 20060302   @24859 Fixed regression related to #005 (@22983).
**                           Natural join information was not properly
**                           persisting in the .schema file, due to AST ID
**                           mismatches between join data (based on .dict
**                           ASTs) and the newer, target, .schema AST.
** 007 ECF 20060325   @25241 Removed code which created duplicate schema
**                           file with the '.original' extension. This
**                           duplication is no longer necessary and is the
**                           source of constant registry problems.
** 008 ECF 20070302   @32265 Ignore self referencing join. That is, a join
**                           between a table and itself. This could happen
**                           if two buffers on the same backing table were
**                           joined using the OF keyword.
** 009 GES 20070629   @34341 Removed terse flag from AST persistence so
**                           that we retail line/column info in our ASTs.
** 010 GES 20070911   @35143 Minor change to how the schema dictionary is
**                           obtained.
** 011 ECF 20080730   @39272 Reduced memory footprint. Implemented finish
**                           method. Made creation of RelationAnalyzer
**                           (which pins SchemaDictionary) lazy. Added a
**                           user function to release SchemaDictionary.
** 012 GES 20081010   @40337 Fix to allow UAST hints to load properly.
** 013 ECF 20081216   @40938 Fixed attachRelationalData(). Allow foreign
**                           relation information to be stored for temp
**                           tables. The only restriction now is that both
**                           tables in the join need to reside in the same
**                           database.
** 014 GES 20090515   @42223 Moved to AstManager from AstRegistry/AstPersister.
** 015 ECF 20130124          Added worker function to get schema config.
** 016 OM  20130808          Added getForeignKey() to obtain the list of the fields used to join
**                           two Progress tables. 
**                           joins Set morphed into a Hashtable for getter fast lookup.
**                           Generified private/local variables except for deprecated methods.
** 017 CA  20140114          Added getLegacyTableName API to resolve the legacy table name 
**                           (case-sensitive) from a table name.
*/

package com.goldencode.p2j.schema;

import java.io.*;
import java.util.*;
import com.goldencode.ast.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.convert.*;
import com.goldencode.p2j.pattern.*;
import com.goldencode.p2j.uast.*;

/**
 * A pattern worker which exposes schema dictionary and table relation
 * analysis functionality to the pattern engine.
 */
public class SchemaWorker
extends AbstractConversionWorker
implements SchemaParserTokenTypes
{
   /** Fixed-up schema AST filename extension. */
   private static final String SCHEMA_POSTFIX = ".schema";
   
   /** List of all Progress field token types */
   private static final ArrayList<Integer> ALL_FIELD_TYPES = new ArrayList<Integer>();
   
   // Initialize list of field token types.
   static
   {
      for (int i = BEGIN_FIELDTYPES + 1; i < END_FIELDTYPES; i++)
      {
         ALL_FIELD_TYPES.add(new Integer(i));
      }
      
      for (int i = BEGIN_METATYPES + 1; i < END_METATYPES; i++)
      {
         ALL_FIELD_TYPES.add(new Integer(i));
      }
      
      ALL_FIELD_TYPES.trimToSize();
   }
   
   /** Schema loader, used to load schema ASTs */
   private SchemaLoader loader = new SchemaLoader();
   
   /** Provides schema information lookup; loads default database(s) */
   private SchemaDictionary dictionary = new SchemaDictionary();
   
   /** Analyzes relations between Progress tables */
   private RelationAnalyzer analyzer = null;
   
   /** Names of databases which contain tables involved in relations */
   private Set<String> databases = new HashSet<String>();
   
   /**
    * Set of all table relations detected in a scanning session. They are hashed using a
    * combination of the table names that assures that there are no access collisions.
    */
   private Map<String, TableRelation> joins = new Hashtable<String, TableRelation>();
   
   /** Set of IDs of index ASTs to be marked as referenced by "use-index" */
   private Set<String> usedIndexes = new HashSet<String>();
   
   /**
    * Default constructor which defines the symbol libraries to be registered.
    *
    * @throws  SchemaException
    *          if there is any error initializing the schema loader or
    *          schema dictionary.
    */
   public SchemaWorker()
   throws SchemaException
   {
      super();
      
      setLibrary(new SchemaSupport());
   }
   
   /**
    * Hook to provide termination processing for a pattern worker.
    */
   public void finish()
   {
      loader = null;
      dictionary = null;
      analyzer = null;
      databases = null;
      joins = null;
      usedIndexes = null;
      
      super.finish();
   }
   
   /**
    * Composes a normalized index name, which is the fully qualified table
    * name in all lowercase, in the form:
    * <code>&lt;database_name&gt;&lt;table_name&gt;&lt;index_name&gt;</code>.
    *
    * @param   index
    *          AST which defines index.
    *
    * @return  Normalized index name as described above.
    */
   private static String normalizedIndexName(Aast index)
   {
      StringBuffer buf = new StringBuffer();
      buf.append(TableRelation.normalizedTableName(index.getParent()))
         .append('.')
         .append(index.getText());
         
      return (buf.toString().toLowerCase());
   }
   
   /**
    * Helper method to append an extension to a filename if it is not
    * already there.
    *
    * @param   filename
    *          Filename which may or may not already end with the extension.
    * @param   extension
    *          Extension to be set at the end of <code>filename</code>.
    *
    * @return  <code>filename</code> with the required extension.
    */
   private String adjustFilename(String filename, String extension)
   {
      if (filename.endsWith(extension))
      {
         return filename;
      }
      
      StringBuffer buf = new StringBuffer(filename);
      buf.append(extension);
      
      return buf.toString();
   }
   
   /**
    * Figures out which schema file to load as the basis for updates. If
    * an original schema file (i.e., one that ends with ".original") exists,
    * it is returned. Otherwise, the file represented by <code>filename</code>
    * is returned.
    *
    * @param   filename
    *          Root schema AST filename.
    *
    * @return  Original filename or base filename if an original has not
    *          been created yet.
    *
    * @throws  ConfigurationException
    *          if P2J_HOME has not been defined.
    */
   private File getSchemaLoadFile(String filename)
   throws ConfigurationException
   {
      return Configuration.toFile(filename);
   }
   
   /**
    * Walks the tables of a database schema AST and attaches relation data
    * ASTs to any table which has been found in a natural join. If any
    * relation data is attached, the specified file is renamed to have an
    * ".original" extension, and the database schema AST is persisted to
    * the original schema filename.
    * <p>
    * The <code>schemaRoot</code> AST will have been loaded from the
    * <code>.schema</code> (or <code>.schema.original</code>) file;  however,
    * all join information is based upon ASTs loaded from the primordial
    * <code>.dict</code> file (needed by the <code>SchemaDictionary</code>).
    * Thus, table IDs within <code>TableRelation</code> objects will not
    * match those in <code>schemaRoot</code> tables.  We must therefore use
    * the normalized table name to associated the table relations with the
    * table ASTs found in the schema AST.  The same technique applies to
    * indexes.
    *
    * @param   schemaRoot
    *          AST which represents a permanent or transient database schema.
    */
   private void persistSchema(Aast schemaRoot, File file)
   throws SchemaException
   {
      // Collect all record-type ASTs, indexing them by ID. Collect all
      // index ASTs, also indexing them by ID.
      Map<String, Aast> tables = new HashMap<String, Aast>();
      Map<String, Aast> indexes = new HashMap<String, Aast>();
      Iterator<Aast> iter = schemaRoot.iterator(2, null);
      while (iter.hasNext())
      {
         Aast next = iter.next();
         switch (next.getType())
         {
            case TABLE:
            case TEMP_TABLE:
            case WORK_TABLE:
               tables.put(TableRelation.normalizedTableName(next), next);
               break;
            
            case INDEX:
               indexes.put(normalizedIndexName(next), next);
               break;
         }
      }
      
      // Attach relational data to tables for which such data has been collected.
      for (TableRelation join : joins.values())
      {
         attachRelationData(join, tables);
      }
      
      // Annotate indexes which are referenced by "use-index".
      Iterator<String> idxIter = usedIndexes.iterator();
      while (idxIter.hasNext())
      {
         Aast index = indexes.get(idxIter.next());
         if (index != null)
         {
            index.putAnnotation("use-index", Boolean.TRUE);
         }
      }
      
      try
      {
         // Persist modified AST.
         String filename = file.getAbsolutePath();
         AstManager.get().saveTree(schemaRoot, filename, false);
      }
      catch (Exception exc)
      {
         throw new SchemaException("Error overwriting schema file", exc);
      }
   }
   
   /**
    * Given a database and record AST, scan the collected join information
    * to see if this record requires table relation information ASTs to be
    * attached. For each join which was recorded in which this table
    * participates on the foreign key end, add an AST describing the join.
    * <p>
    * The ASTs which are attached are of token type ONE_TO_ONE or MANY_TO_ONE,
    * depending upon the circumstances of the join. Each has one child AST
    * of type KEY_FIELD for each field which participates in the join. In
    * addition, each has an annotation to record the ID of the table on the
    * other end of the join. The ASTs are attached as direct children of the
    * table to which they refer.
    *
    * @param   join
    *          Object which describes a natural relation between tables.
    * @param   tables
    *          Map of record-type AST nodes, indexed by ID. This map may or
    *          may not contain one or both tables described by
    *          <code>join</code>.
    *
    * @throws  SchemaException
    *          if any error occurs attaching a join information AST to
    *          the table AST <code>ast</code>.
    */
   private void attachRelationData(TableRelation join, Map<String, Aast> tables)
   throws SchemaException
   {
      String key;
      
      // Get the node within the AST to be updated, which represents the
      // left side of this join. If this join has a primary end, it will be
      // the left table.
      key = TableRelation.normalizedTableName(join.getLeft());
      Aast left = tables.get(key);
      
      // Get the node within the AST to be updated, which represents the
      // right side of this join. If this join has a foreign end, it will be
      // the right table.
      key = TableRelation.normalizedTableName(join.getRight());
      Aast right = tables.get(key);
      
      // Graft join data onto each table iff it exists in the current database
      // and its inverse also exists in the same database.
      if (left != null && right != null)
      {
         graftJoinNode(join.getInverseType(), join.getKey(), left, right);
         graftJoinNode(join.getType(), join.getKey(), right, left);
      }
   }
   
   /**
    * Graft an AST node which describes a natural join onto a target table
    * AST. The node's token type describes the nature of the join. It has
    * an annotation with the ID of the table on the inverse end of the
    * relation. The node has one child per field involved in the key which
    * binds the tables.
    *
    * @param   type
    *          Token type of the join node.
    * @param   key
    *          Set of names of fields which form the key between
    *          <code>target</code> and <code>inverse</code>.
    * @param   target
    *          Table AST to which join node is added.
    * @param   inverse
    *          AST which represents the table on the inverse end of the
    *          relation.
    *
    * @throws  SchemaException
    *          if an error occurs adding the join AST node.
    */
   private void graftJoinNode(int type, Set key, Aast target, Aast inverse)
   throws SchemaException
   {
      String text = TableRelation.normalizedTableName(inverse);
      AstSymbolResolver resolver = AstSymbolResolver.getResolver();
      
      // Create relation AST child to graft to the target table node.
      Aast relAst = new ProgressAst();
      initializeAst(relAst, type, text, target, resolver);
      
      // Add an annotation indicating the table on the other end of the
      // relation.
      relAst.putAnnotation("references", inverse.getId());
      
      // Add each field as a child component of the relation node.
      Iterator<String> fieldIter = key.iterator();
      while (fieldIter.hasNext())
      {
         String fieldName = fieldIter.next();
         
         Aast compAst = new ProgressAst();
         initializeAst(compAst, KEY_FIELD, fieldName, relAst, resolver);
         
         // Annotate with a reference to actual, local field.
         Aast field = target.getImmediateChild(ALL_FIELD_TYPES, null);
         while (field != null)
         {
            if (fieldName.equals(field.getText()))
            {
               compAst.putAnnotation("refid", field.getId());
               break;
            }
            field = target.getImmediateChild(ALL_FIELD_TYPES, field);
         }
         
         // Annotate with a cross-reference to same field in inverse table.
         field = inverse.getImmediateChild(ALL_FIELD_TYPES, null);
         while (field != null)
         {
            if (fieldName.equals(field.getText()))
            {
               compAst.putAnnotation("crossrefid", field.getId());
               break;
            }
            field = inverse.getImmediateChild(ALL_FIELD_TYPES, field);
         }
      }
   }
   
   /**
    * Provides schema-related services to pattern engine users. An object of
    * this class is registered with the symbol resolver to provide user
    * function support for user expressions.
    */
   public class SchemaSupport
   {
      private List redundantIndexReport = new ArrayList();
      
      private List indexComponentReport = new ArrayList();
      
      /**
       * Return instance of schema dictionary.
       *
       * @return  Current instance of schema dictionary.
       */
      public SchemaDictionary getDictionary()
      {
         return dictionary;
      }
      
      /**
       * Release the in-use dictionary.
       */
      public void releaseDictionary()
      {
         dictionary = null;
      }
      
      /**
       * Resolve the legacy table name, given the (possible) qualified table name.
       * 
       * @param    name
       *           The (possible) qualified table name.
       * 
       * @return   See above.
       */
      public String getLegacyTableName(String name)
      {
         try
         {
            Aast table = dictionary.getTable(name);
            
            String legacyName = (String) table.getAnnotation("legacy_name");
            if (legacyName == null)
            {
               throw new SchemaException("Legacy name not found for table:" + name);
            }
            
            return legacyName;
         }
         catch (SchemaException e)
         {
            throw new RuntimeException("Could not resolve legacy name for table: " + name);
         }
      }
      
      /**
       * Initialize the schema dictionary for use with the current source
       * file AST. Resets the state of the dictionary, but keeps default
       * databases loaded. Also checks for hints to see if any additional
       * databases need to be loaded and/or aliases assigned.
       *
       * @param   filename
       *          Name of the source file represented by the AST currently
       *          being processed in the pattern engine. Any temp- and work-
       *          tables defined in this source file are loaded into the
       *          dictionary.
       *
       * @return  <code>true</code>.
       *
       * @throws  SchemaException
       *          if any error occurs loading source file-specific schema
       *          information.
       */
      public boolean initializeDictionary(String filename)
      throws SchemaException
      {
         // Get a clean instance of the dictionary.
         dictionary = new SchemaDictionary();
         
         // Load hints for this source file, if any.
         UastHints hints = new UastHints(filename);
         
         // Load extra databases into dictionary, if any.
         String[] dbList = hints.getDatabases();
         if (dbList != null)
         {
            for (int i = 0; i < dbList.length; i++)
            {
               dictionary.loadSchema(dbList[i]);
            }
         }
         
         // Create database aliases, if any.
         List<String[]> aliases = hints.getAliases();
         if (aliases != null)
         {
            Iterator<String[]> iter = aliases.iterator();
            while (iter.hasNext())
            {
               String[] pair = iter.next();
               
               if (!dictionary.createAlias(pair[0], pair[1]))
               {
                  System.err.println(
                     "WARNING:  dictionary could not create alias " +
                     pair[0] + " for database " + pair[1]);
               }
            }
         }
         
         // Load AST which defines temp- and work-tables for the current
         // source file.
         filename = adjustFilename(filename, AstGenerator.DICT_POSTFIX);
         File file = new File(filename);
         if (file.exists())
         {
            dictionary.loadFromPersistence(file.getAbsolutePath());
         }
         
         return true;
      }
      
      /**
       * Get the {@link SchemaConfig} that was read from the project
       * configuration.
       * 
       * @return  Schema configuration.
       */
      public SchemaConfig getSchemaConfig()
      {
         return Configuration.getSchemaConfig();
      }
      
      /**
       * Return a string which describes a join between two tables. This is
       * useful for debug and ad-hoc reporting purposes.
       *
       * @param   left
       *          Left-hand table of the join (fully qualified).
       * @param   right
       *          Right-hand table of the join (fully qualified).
       *
       * @return  A string describing the relation between these tables,
       *          if any.  <code>null</code> if there was an error.
       */
      public String describeJoin(String left, String right)
      {
         // Ignore a table which is joined to itself.
         if (left.equals(right))
         {
            return null;
         }
         
         String description = null;
         
         try
         {
            if (analyzer == null)
            {
               analyzer = new RelationAnalyzer(dictionary);
            }
            TableRelation join = analyzer.createJoin(left, right);
            description = join.toString();
         }
         catch (SchemaException exc)
         {
            description = exc.getMessage();
         }
         catch (Exception exc)
         {
            exc.printStackTrace();
         }
         
         return description;
      }
      
      /**
       * Create a {@link TableRelation} object based the natural join of
       * the specified tables, and add it to the set of joins being collected
       * by this worker. Error messages are printed to <code>stderr</code>.
       *
       * @param   left
       *          Left-hand table of the join (fully qualified).
       * @param   right
       *          Right-hand table of the join (fully qualified).
       *
       * @return  The instance of a <code>TableRelation</code>, if the join
       *          was successfully recorded;  <code>null</code> if there was
       *          an error.
       */
      public TableRelation recordJoin(String left, String right)
      {
         // Ignore a table which is joined to itself.
         if (left.equals(right))
         {
            return null;
         }
         
         /* 
          * OM: The TableRelation is reflexive from Progress point of view. This is different from
          * normal SQL notion of table join by foreign key. 
          * The Progress imposes that:
          * - the name of the fields that compose the index to have the same name in both tables,
          *       if the relation is specified all fields participate in the join implicitly.
          * - such combination to be unique.
          * The analyzer will automatically detect the unique indexed table and choose it as the
          * sql left side of the join.
          * Because of that, createJoin(left, right) and createJoin(right, left) will generate
          * "equals" objects, so we can freely reorder the tables.  If we index the "joins"
          * map using the alphabetic order of the table names we can quickly lookup the relation
          * between the two tables.
          */
         String joinLookup = left.compareTo(right) > 0 ? right + " " + left : left + " " + right;
         TableRelation join = joins.get(joinLookup);
         if (join != null)
         {
            // return the cached relation between the two tables
            return join;
         }

         try
         {
            if (analyzer == null)
            {
               analyzer = new RelationAnalyzer(dictionary);
            }
            join = analyzer.createJoin(left, right);
            
            /*
             * Save the relation for later lookups. It will:
             * - cache the value for future calls for same tables (which can be swapped)
             * - allow extracting the fields of key for nested can-find cases where the relation
             *       between the two tables has to be expanded explicitly in hql.  
             */
            joins.put(joinLookup, join);
            databases.addAll(join.getDatabases());
         }
         catch (SchemaException exc)
         {
            System.err.println(exc.getMessage());
         }
         catch (Exception exc)
         {
            exc.printStackTrace();
         }
         
         return join;
      }
      
      /**
       * Return a string which contains the fields (space separated) that are part of the foreign
       * key that joins two tables. The relation should already been added with 
       * {@link #recordJoin(String, String)}.
       *
       * @param   left
       *          Left-hand table of the join (fully qualified).
       * @param   right
       *          Right-hand table of the join (fully qualified).
       *
       * @return  A string describing the foreign key of these tables, if any, and 
       *          <code>null</code> if tables are unrelated or there was an error.
       */
      public String getForeignKey(String left, String right)
      {
         // Eliminate common cases
         if (left == null || right == null || left.equals(right))
         {
            return null;
         }
         
         String joinLookup = left.compareTo(right) > 0 ? right + " " + left : left + " " + right;
         TableRelation join = joins.get(joinLookup);
         if (join == null)
         {
            return null;
         }
         
         StringBuffer sb = new StringBuffer();
         for (Object k : join.getKey())
         {
            if (sb.length() > 0)
            {
               sb.append(' ');
            }
            sb.append((String)k);
         }
         return sb.toString();
      }
      
      /**
       * Add index AST to the set of indexes referenced by the Progress
       * "use-index" directive. Update the set of databases to be persisted
       * as a result of this discovery. The actual annotation occurs later,
       * when the corresponding database AST is persisted.
       *
       * @param   index
       *          AST of the index to be marked.
       */
      public void markIndexUsed(Aast index)
      {
         usedIndexes.add(normalizedIndexName(index));
         databases.add(index.getAncestor(-1).getText());
      }
      
      /**
       * Persist the modified Progress schema ASTs which represent permanent
       * database tables. These ASTs are now updated with table relation
       * information. For each AST so modified, rename its original file
       * with a new extension, and re-use the previous name to persist this
       * modified AST.
       *
       * @return  <code>true</code> if method complets without exception.
       *
       * @throws  ConfigurationException
       *          if P2J_HOME environment variable is not set.
       * @throws  SchemaException
       *          if there were errors updating or saving the affected
       *          schema ASTs.
       */
      public boolean persistPermanent()
      throws ConfigurationException,
             SchemaException
      {
         Iterator<String> iter = databases.iterator();
         while (iter.hasNext())
         {
            String next = iter.next();
            if (next.startsWith("@"))
            {
               continue;
            }
            
            File saveFile = loader.getSchemaFile(next);
            String filename = saveFile.getAbsolutePath();
            filename = filename.substring(0, filename.lastIndexOf("."));
            filename += SCHEMA_POSTFIX;
            saveFile = new File(filename);
            File loadFile = getSchemaLoadFile(filename);
            if (loadFile.exists())
            {
               persistSchema(loader.loadAst(loadFile), saveFile);
            }
         }
         
         return true;
      }
      
      /**
       * Persist the AST for all temp- and work-tables associated with the
       * given source file, if this AST has been updated with any join data.
       * This AST now contains foreign key information. Store it using a
       * separate file extension, so as not to overwrite the original
       * schema file.
       * <p>
       * If <code>filename</code> is <code>null</code>, this indicates that
       * all permanent database ASTs which have been loaded should be saved.
       */
      public boolean persistTransient(String filename)
      throws ConfigurationException,
             SchemaException
      {
         filename = adjustFilename(filename, SCHEMA_POSTFIX);
         File loadFile = getSchemaLoadFile(filename);
         if (loadFile.exists())
         {
            File saveFile = Configuration.toFile(filename);
            persistSchema(loader.loadAst(loadFile), saveFile);
         }
         
         return true;
      }
      
      /**
       * Dump to <code>stdout</code> the set of natural joins detected by
       * all invocations of {@link #recordJoin} during a scan of some set
       * of Progress source code.
       */
      public void dumpNaturalJoins()
      {
         System.out.println("ALL NATURAL JOINS: ");
         System.out.println("------------------");
         for (TableRelation join : joins.values())
         {
            System.out.println(join.toString());
         }
      }
      
      /**
       * @deprecated
       */
      public void analyzeIndexComponents(String table, Map indices)
      {
         List report = new ArrayList();
         Set<String> done = new HashSet<String>();
         
         Iterator iter = indices.keySet().iterator();
         while (iter.hasNext())
         {
            // Get current index name.
            String name = (String) iter.next();
            done.add(name);
            
            // Get list of current index component fields.
            List fields = (List) indices.get(name);
            
            Iterator iter2 = indices.keySet().iterator();
            while (iter2.hasNext())
            {
               String key = (String) iter2.next();
               
               // Skip indices already reviewed.
               if (done.contains(key))
               {
                  continue;
               }
               
               List fields2 = (List) indices.get(key);
               
               // Iterate through "outer" index' fields.
               boolean problem = false;
               Iterator subIter = fields.iterator();
               for (int i = 0; subIter.hasNext(); i++)
               {
                  String next = (String) subIter.next();
                  int index = fields2.indexOf(next);
                  if (index >= 0 && index != i)
                  {
                     if (!problem)
                     {
                        problem = true;
                        report.add("   " + name + ":  " + fields);
                        report.add("   " + key + ":  " + fields2);
                     }
                     report.add("      " + next + ":  " + name + "[" + i +
                                "] ... " + key + "[" + index + "]");
                  }
               }
            }
         }
         
         if (report.size() > 0)
         {
            indexComponentReport.add(table + ":");
            iter = report.iterator();
            while (iter.hasNext())
            {
               indexComponentReport.add(iter.next());
            }
         }
      }
      
      /**
       * @deprecated
       */
      public void analyzeIndexRedundancy(String table, Map indices)
      {
         List report = new ArrayList();
         Set<String> done = new HashSet<String>();
         
         Iterator iter = indices.keySet().iterator();
         while (iter.hasNext())
         {
            // Get current index name.
            String name = (String) iter.next();
            done.add(name);
            
            // Get list of current index component fields.
            List fields = (List) indices.get(name);
            int size = fields.size();
            
            Iterator iter2 = indices.keySet().iterator();
            while (iter2.hasNext())
            {
               String key = (String) iter2.next();
               
               // Skip indices already reviewed.
               if (done.contains(key))
               {
                  continue;
               }
               
               // Get list of compare target's component fields.
               List tgtFields = (List) indices.get(key);
               int tgtSize = tgtFields.size();
               
               String nameA;
               String nameB;
               List listA;
               List listB;
               
               if (size < tgtSize)
               {
                  nameA = name;
                  nameB = key;
                  listA = fields;
                  listB = tgtFields;
               }
               else
               {
                  nameA = key;
                  nameB = name;
                  listA = tgtFields;
                  listB = fields;
               }
               
               // Compare first N components of larger list with all
               // components of smaller list.
               if (listA.equals(listB.subList(0, listA.size())))
               {
                  report.add("   " + nameA + " is redundant with " + nameB +
                             ":");
                  report.add("      " + listA);
                  report.add("      " + listB);
               }
            }
         }
         
         if (report.size() > 0)
         {
            redundantIndexReport.add(table + ":");
            iter = report.iterator();
            while (iter.hasNext())
            {
               redundantIndexReport.add(iter.next());
            }
         }
      }
      
      /**
       * @deprecated
       */
      public void generateIndexReports(String filename)
      throws IOException
      {
         BufferedWriter bw = new BufferedWriter(new FileWriter(filename));
         
         try
         {
            bw.write("REDUNDANT INDICES:");
            bw.newLine();
            bw.write("------------------");
            bw.newLine();
            
            if (redundantIndexReport.size() == 0)
            {
               bw.write("none");
               bw.newLine();
               bw.newLine();
            }
            else
            {
               Iterator iter = redundantIndexReport.iterator();
               while (iter.hasNext())
               {
                  bw.write((String) iter.next());
                  bw.newLine();
               }
               bw.newLine();
            }
            
            bw.write("PROBLEM INDEX COMPONENTS:");
            bw.newLine();
            bw.write("-------------------------");
            bw.newLine();
            
            if (indexComponentReport.size() == 0)
            {
               bw.write("none");
               bw.newLine();
               bw.newLine();
            }
            else
            {
               Iterator iter = indexComponentReport.iterator();
               while (iter.hasNext())
               {
                  bw.write((String) iter.next());
                  bw.newLine();
               }
               bw.newLine();
            }
         }
         finally
         {
            if (bw != null)
            {
               bw.close();
            }
         }
      }
   }
}
